typedef char buf_char_t;

/* color attribute data */
typedef struct {
	/* bitfields instead of attr flags
	 * not needed in vga compatible mode */
	unsigned int bold : 1; // included in the color ?
	unsigned int halfbright : 1;
	unsigned int underline : 1;
	unsigned int blink : 1;
	unsigned int reverse : 1; // included in the color ?
} buf_attr_t;

/* color data */
typedef struct {
	unsigned char fg;	/* foreground color */
	unsigned char bg;	/* background color */
	buf_attr_t attr;	/* color attributes */
} buf_color_t;

/* character cell data */
typedef struct {
	buf_char_t c;		/* character */
	buf_color_t color;	/* color */
} buf_cell_t;

/* cursor type */
typedef enum {
	buf_cursor_default,
	buf_cursor_none,
	buf_cursor_lower_third,
	buf_cursor_lower_half,
	buf_cursor_two_thirds,
	buf_cursor_block
} buf_cursor_t;

/* common buffer state */
typedef struct {
	int cursor_x;	/* cursor position */
	int cursor_y;	/* cursor position */

	int w;			/* (visible) buffer width */
	int h;			/* (visible) buffer height */
	
	int scrv;		/* scroll amount */

	buf_color_t color;
	buf_color_t default_color;
	unsigned char underline_fg;
	unsigned char halfbright_fg;
	
	buf_cursor_t cursor_type;
} buf_state_t;

#ifndef HAVE_BUFFER

/* display->buffer: creates new buffer, returns handle.
 * buffer will create emu and io */
void *buf_create();

/* display->buffer: set screen dimensions (visible buffer size) */
void buf_set_screen_dim(const int w, const int h);

/* set internal r/w pointer in (visible) buffer */
void buf_set_rw(void *b, const int x, const int y);

/* set internal r/w pointer in all buffer (with scroll support, only for viewing) */
void buf_set_r(void *b, const int x, const int y);

/* move rw to the cursor position */
static inline void buf_rw_from_cursor(void *b);

/* reads next character cell from (visible) buffer, advances r/w pointer
 * returns buf_cell_t
 */
buf_cell_t buf_read_cell(void *b);

/* writes character (buf_char_t) into buffer, advances r/w pointer
 * uses current colors/attributes
 */
void buf_write_char(void *b, buf_char_t ch);

/* returns pointer to state of current buffer
 * buf_state_t *
 * don't write */
buf_state_t * buf_get_state(void *b);

/* set cursor position. sets r/w pointer to the same position */
void buf_set_cursor(void *b, const int x, const int y);

/* move cursor and r/w left by one position */
static inline void buf_cursor_left(void *b);

/* move cursor and r/w right by one position */
static inline void buf_cursor_right(void *b);

/* set foreground color */
void buf_set_fg(void *b, unsigned char fg);

/* set background color */
void buf_set_bg(void *b, unsigned char bg);

/* set attrs */
void buf_set_attr(void *b, unsigned char attr);

/* set color */
void buf_set_color(void *b, buf_color_t color);

/* set color to default color */
void buf_reset_color(void *b);

/* scroll the buffer. bottom line will be filled with 0, default fg/bg/attr */
void buf_scroll(void *b);

/* process multiple key press events (sends to emu) */
static inline void buf_dsp_send(void *b, char *s);

/* process key press event (sends to emu) */
void buf_dsp_char(void *b, char c);

/* process ctrl event (sends to emu) */
void buf_dsp_ctrl(void *b, ctrl_t ctrl);

/* invalidate display */
void buf_update(void *b, const int x1, const int y1, const int x2, const int y2);

/* hide cursor */
void buf_cursor_hide(void *b);

/* select active buffer */
void buf_select(buf_t *b);

/* adjust buffer scroll */
void buf_adjust_scroll(buf_t *b, int amount);

/* switch buffer relatively */
void * buf_switch_rel(const int o);

/* death callback of buffer */
void buf_dead(void *b);

/* erase part of buffer */
void buf_erase(void *b, const int x1, const int y1, const int x2, const int y2);

/* shift part of buffer */
void buf_shift(void *b, const int x1, const int y1, const int x2, const int y2, const int xd, const int yd);

#error No BUFFER module is selected.

#endif /* HAVE_BUFFER */
