#ifndef buffer_real_h
#define buffer_real_h

#ifdef HAVE_BUFFER
#error Can not have more than one BUFFER module.
#endif

#define HAVE_BUFFER
#include "buffer/interface.h"
#include DISPLAY
#include EMULATION
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include "global.h"

#ifndef CFG_BUFFERS_MAX
#	define CFG_BUFFERS_MAX 32
#endif

#ifndef CFG_BUFFER_SCROLLBACK
#	define CFG_BUFFER_SCROLLBACK 20
#endif

#define MAXBUFFERS CFG_BUFFERS_MAX

/* default color and attrs */
#define DEFAULT_ATTR (buf_attr_t){.bold = 0, .halfbright = 0, .underline = 0, .blink = 0, .reverse = 0}
#define DEFAULT_COLOR (buf_color_t){.fg = 7, .bg = 0, .attr = DEFAULT_ATTR}

/* VGA emulates these attrs with foreground color */
#define UNDERLINE_FG 15 /* underline foreground color */
#define HALFBRIGHT_FG 7 /* halfbright foreground color */

typedef struct {
	buf_cell_t **bufdata;
	buf_cell_t *rwptr;
	buf_state_t bufstate;
	void *bufemu;
	int loff; /* offset of line 0 */
	int bufoff;
} buf_t;

extern int bufw, bufh, scrh, bufn;

extern buf_t buffers[MAXBUFFERS];
extern buf_t *cb;

/* invalidate display */
void buf_update(buf_t *b, const int x1, const int y1, const int x2, const int y2);

/* hide cursor */
static inline void buf_cursor_hide(buf_t *b)
{
	if(b == cb)	display_cursor_hide();
}

/* set internal r/w pointer in (visible) buffer */
static inline void buf_set_rw(buf_t *b, const int x, const int y)
{
	int o = (b->bufoff + y) % bufh;
	if(!b->bufdata[o]) {
		b->bufdata[o] = calloc(bufw, sizeof(buf_cell_t));
		assert(b->bufdata[o] != NULL);
	}

	b->rwptr = b->bufdata[o] + x;
}

/* set internal r/w pointer in all buffer (with scroll support, only for viewing) */
static inline void buf_set_r(buf_t *b, const int x, const int y)
{
	int o = (b->bufoff + y - b->bufstate.scrv) % bufh;
	if(!b->bufdata[o]) {
		b->bufdata[o] = calloc(bufw, sizeof(buf_cell_t));
		assert(b->bufdata[o] != NULL);
	}

	b->rwptr = b->bufdata[o] + x;
}

/* actualize buffer offset */
static inline void buf_offact(buf_t *b)
{
	b->bufoff = b->loff + bufh - scrh;
}

/* set cursor position. sets r/w pointer to the same position */
static inline void buf_set_cursor(buf_t *b, const int x, const int y)
{
	buf_cursor_hide(b);

	b->bufstate.cursor_x = x;
	b->bufstate.cursor_y = y;
	buf_set_rw(b, x, y);

	buf_update(b, x, y, x, y);
}

void *buf_create();

/* display->buffer: set screen dimensions (visible buffer size) */
void buf_set_screen_dim(const int w, const int h);

/* move rw to the cursor position */
static inline void buf_rw_from_cursor(buf_t *b)
{
	buf_set_rw(b, b->bufstate.cursor_x, b->bufstate.cursor_y);
}

/* move r/w left by one position */
static inline void buf_rw_left(buf_t *b)
{
	b->rwptr--;
}

/* move r/w right by one position */
static inline void buf_rw_right(buf_t *b)
{
	b->rwptr++;
}

/* reads next character cell from (visible) buffer, advances r/w pointer
 * returns buf_cell_t
 */
static inline buf_cell_t buf_read_cell(buf_t *b)
{
	return *(b->rwptr++);
}

/* writes character (buf_char_t) into buffer, advances r/w pointer
 * uses current colors/attributes
 */
static inline void buf_write_char(buf_t *b, buf_char_t ch)
{
	b->rwptr->c = ch;
	b->rwptr->color = b->bufstate.color;
	b->rwptr++;
}

/* returns pointer to state of current buffer
 * buf_state_t *
 * don't write */
static inline buf_state_t * buf_get_state(buf_t *b)
{
	return &b->bufstate;
}

/* move cursor and r/w left by one position */
static inline void buf_cursor_left(buf_t *b)
{
	buf_cursor_hide(b);
	b->bufstate.cursor_x--;
	display_update(b->bufstate.cursor_x, b->bufstate.cursor_y, b->bufstate.cursor_x, b->bufstate.cursor_y);
	buf_rw_left(b);
}

/* move cursor and r/w right by one position */
static inline void buf_cursor_right(buf_t *b)
{
	b->bufstate.cursor_x++;
	buf_rw_right(b);
}

/* move cursor and r/w up by one position */
static inline void buf_cursor_up(buf_t *b)
{
	buf_cursor_hide(b);
	b->bufstate.cursor_y--;
	if(b->bufstate.cursor_y < 0) b->bufstate.cursor_y += bufh;
	buf_rw_from_cursor(b);
}

/* move cursor and r/w down by one position */
static inline void buf_cursor_down(buf_t *b)
{
	buf_cursor_hide(b);
	b->bufstate.cursor_y++;
	if(b->bufstate.cursor_y == bufh) b->bufstate.cursor_y = 0;
	buf_rw_from_cursor(b);
}

/* set update color settings */
static inline void buf_update_color(buf_t *b)
{
}

/* set background color */
static inline void buf_set_fg(buf_t *b, unsigned char fg)
{
	b->bufstate.color.fg = fg;
	buf_update_color(b);
}

/* set background color */
static inline void buf_set_bg(buf_t *b, unsigned char bg)
{
	b->bufstate.color.bg = bg;
	buf_update_color(b);
}

/* set attrs */
static inline void buf_set_attr(buf_t *b, buf_attr_t attr)
{
	b->bufstate.color.attr = attr;
	buf_update_color(b);
}

/* set color */
static inline void buf_set_color(buf_t *b, buf_color_t color)
{
	b->bufstate.color = color;
	buf_update_color(b);
}

/* set color to default color */
static inline void buf_reset_color(buf_t *b)
{
	b->bufstate.color =
	b->bufstate.default_color = DEFAULT_COLOR;
	b->bufstate.underline_fg = UNDERLINE_FG;
	b->bufstate.halfbright_fg = HALFBRIGHT_FG;
}

/* process multiple key press events (sends to emu) */
static inline void buf_dsp_send(buf_t *b, char *s)
{
	char *p;
	for(p = s; *p; p++)
		emu_dsp_char(b->bufemu, *p);
}

/* process key press event (sends to emu) */
static inline void buf_dsp_char(buf_t *b, char c)
{
	emu_dsp_char(b->bufemu, c);
}

/* process ctrl key press event (sends to emu) */
static inline void buf_dsp_ctrl(buf_t *b, ctrl_t c)
{
	emu_dsp_ctrl(b->bufemu, c);
}

/* select active buffer */
void buf_select(buf_t *b);

/* scroll the buffer. bottom line will be filled with 0, default fg/bg/attr */
void buf_scroll(buf_t *b);

/* adjust buffer scroll */
void buf_adjust_scroll(buf_t *b, int amount);

/* switch buffer relatively */
void * buf_switch_rel(const int o);

/* death callback of buffer */
void buf_dead(buf_t *b);

/* shift part of buffer */
void buf_shift(buf_t *b, const int x1, const int y1, const int x2, const int y2, const int xd, const int yd);

/* erase part of buffer */
void buf_erase(buf_t *b, const int x1, const int y1, const int x2, const int y2);

#endif
